#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################


def abs_energy(arr):
    """ Calculates the sum over the square values of the time series.

    :param arr: KHIVA array with the time series.
    :type arr: khiva.array
    :return: KHIVA array with the absEnergy.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.abs_energy(ctypes.pointer(arr.arr_reference),
                                              ctypes.pointer(b))
    return Array(array_reference=b)


def absolute_sum_of_changes(arr):
    """ Calculates the value of an aggregation function f_agg (e.g. var or mean) of the autocorrelation
    (Compare to http://en.wikipedia.org/wiki/Autocorrelation#Estimation), taken over different all possible
    lags (1 to length of x)

    :param arr: KHIVA array with the time series.
    :type arr: khiva.array
    :return: KHIVA array with the absolute sum of changes.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.absolute_sum_of_changes(ctypes.pointer(arr.arr_reference),
                                                           ctypes.pointer(b))
    return Array(array_reference=b)


def aggregated_autocorrelation(arr, aggregation_function):
    """ Calculates a linear least-squares regression for values of the time series that were aggregated
    over chunks versus the sequence from 0 up to the number of chunks minus one.

    :param arr: A KHIVA array with the time series.
    :param aggregation_function: Function to be used in the aggregation. It receives an integer which indicates
                                the function to be applied.
                                0 : mean,
                                1 : median
                                2 : min,
                                3 : max,
                                4 : stdev,
                                5 : var,
                                default : mean

    :return: KHIVA array that contains the aggregated correlation for each time series.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.aggregated_autocorrelation(ctypes.pointer(arr.arr_reference),
                                                              ctypes.pointer(ctypes.c_int(aggregation_function)),
                                                              ctypes.pointer(b))

    return Array(array_reference=b)


def aggregated_linear_trend(arr, chunk_size, aggregation_function):
    """ Calculates a linear least-squares regression for values of the time series that were aggregated
    over chunks versus the sequence from 0 up to the number of chunks minus one.

    :param arr: A KHIVA array with the time series.
    :param chunk_size: The chunk size used to aggregate the data.
    :param aggregation_function: Function to be used in the aggregation. It receives an integer which indicates the
                                function to be applied:
                                0 : mean,
                                1 : median
                                2 : min,
                                3 : max,
                                4 : stdev,
                                default : mean
    :return: ( pvalue: KHIVA array with the pvalues for all time series.
            rvalue: KHIVA array with the rvalues for all time series.
            intercept: KHIVA array with the intercept values for all time series.
            slope: KHIVA array with the slope for all time series.
            stdrr: KHIVA array with the stderr values for all time series. )
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)
    e = ctypes.c_void_p(0)
    f = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.aggregated_linear_trend(ctypes.pointer(arr.arr_reference),
                                                           ctypes.pointer(ctypes.c_long(chunk_size)),
                                                           ctypes.pointer(ctypes.c_int(aggregation_function)),
                                                           ctypes.pointer(b),
                                                           ctypes.pointer(c),
                                                           ctypes.pointer(d),
                                                           ctypes.pointer(e),
                                                           ctypes.pointer(f))

    return Array(array_reference=b), Array(array_reference=c,
                                           khiva_type=arr.khiva_type), Array(
        array_reference=d), Array(array_reference=e), Array(
        array_reference=f)


def approximate_entropy(arr, m, r):
    """ Calculates a vectorized Approximate entropy algorithm.
    https://en.wikipedia.org/wiki/Approximate_entropy
    For short time-series this method is highly dependent on the parameters, but should be stable for N > 2000,
    see: Yentes et al. (2012) - The Appropriate Use of Approximate Entropy and Sample Entropy with Short Data Sets
    Other shortcomings and alternatives discussed in:
    Richman & Moorman (2000) - Physiological time-series analysis using approximate entropy and sample entropy


    :param arr: A KHIVA array with the time series.
    :param m: Length of compared run of data.
    :param r: Filtering level, must be positive.
    :return: KHIVA array with the vectorized approximate entropy for all the input time series in tss.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.approximate_entropy(ctypes.pointer(arr.arr_reference),
                                                       ctypes.pointer(ctypes.c_int(m)),
                                                       ctypes.pointer(ctypes.c_float(r)),
                                                       ctypes.pointer(b))

    return Array(array_reference=b)


def cross_covariance(xss, yss, unbiased):
    """ Calculates the cross-covariance of the given time series.

    :param xss: A KHIVA array with time series.
    :param yss: A KHIVA Array with time series.
    :param unbiased: Determines whether it divides by n - lag (if true) or n (if false).
    :return: KHIVA array with the cross-covariance value for the given time series.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.cross_covariance(ctypes.pointer(xss.arr_reference),
                                                    ctypes.pointer(yss.arr_reference),
                                                    ctypes.pointer(ctypes.c_bool(unbiased)), ctypes.pointer(b))

    return Array(array_reference=b)


def auto_covariance(arr, unbiased=False):
    """ Calculates the auto-covariance the given time series.

    :param arr: KHIVA array with the time series.
    :param unbiased: Determines whether it divides by n - lag (if true) or n (if false).
    :return: KHIVA array with the auto-covariance value for the given time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.auto_covariance(ctypes.pointer(arr.arr_reference),
                                                   ctypes.pointer(ctypes.c_bool(unbiased)),
                                                   ctypes.pointer(b))

    return Array(array_reference=b)


def cross_correlation(xss, yss, unbiased):
    """ Calculates the cross-correlation of the given time series.

    :param xss: KHIVA array with the time series.
    :param yss: KHIVA array with the time series.
    :param unbiased: Determines whether it divides by n - lag (if true) or n (if false).
    :return: KHIVA array with cross-correlation value for the given time series.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.cross_correlation(ctypes.pointer(xss.arr_reference),
                                                     ctypes.pointer(yss.arr_reference),
                                                     ctypes.pointer(ctypes.c_bool(unbiased)), ctypes.pointer(b))

    return Array(array_reference=b)


def auto_correlation(arr, max_lag, unbiased):
    """ Calculates the autocorrelation of the specified lag for the given time series.

    :param arr: KHIVA array with the time series.
    :param max_lag: The maximum lag to compute.
    :param unbiased: Determines whether it divides by n - lag (if true) or n (if false).
    :return: KHIVA array with the autocorrelation value for the given time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.auto_correlation(ctypes.pointer(arr.arr_reference),
                                                    ctypes.pointer(ctypes.c_long(max_lag)),
                                                    ctypes.pointer(ctypes.c_bool(unbiased)),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def binned_entropy(arr, max_bins):
    """ Calculates the binned entropy for the given time series and number of bins.

    :param arr: KHIVA array with the time series.
    :param max_bins: The number of bins.
    :return: KHIVA array with the binned entropy value for the given time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.binned_entropy(ctypes.pointer(arr.arr_reference),
                                                  ctypes.pointer(ctypes.c_int(max_bins)),
                                                  ctypes.pointer(b))

    return Array(array_reference=b)


def c3(arr, lag):
    """ Calculates the Schreiber, T. and Schmitz, A. (1997) measure of non-linearity
    for the given time series

    :param arr: KHIVA array with the time series.
    :param lag: The lag.
    :return: KHIVA array with non-linearity value for the given time series.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.c3(ctypes.pointer(arr.arr_reference),
                                      ctypes.pointer(ctypes.c_long(lag)), ctypes.pointer(b))

    return Array(array_reference=b)


def cid_ce(arr, z_normalize):
    """ Calculates an estimate for the time series complexity defined by
    Batista, Gustavo EAPA, et al (2014). (A more complex time series has more peaks,
    valleys, etc.)

    :param arr: KHIVA array with the time series.
    :param z_normalize: Controls wheter the time series should be z-normalized or not.
    :return: KHIVA array with the complexity value for the given time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.cid_ce(ctypes.pointer(arr.arr_reference),
                                          ctypes.pointer(ctypes.c_bool(z_normalize)), ctypes.pointer(b))

    return Array(array_reference=b)


def count_above_mean(arr):
    """ Calculates the number of values in the time series that are higher than
    the mean.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the number of values in the time series that are higher than the mean.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.count_above_mean(ctypes.pointer(arr.arr_reference),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def count_below_mean(arr):
    """ Calculates the number of values in the time series that are lower than
    the mean.

    :param arr: KHIVA array with the time series.
    :return: KHIVA Array with the number of values in the time series that are lower than the mean.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.count_below_mean(ctypes.pointer(arr.arr_reference),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def cwt_coefficients(tss, widths, coeff, w):
    """ Calculates a Continuous wavelet transform for the Ricker wavelet, also known as
    the "Mexican hat wavelet".

    :param tss: KHIVA array with the time series.
    :param widths: Widths. It accepts a list of lists or a numpy array with one or several widths.
    :param coeff: Coefficient of interest.
    :param w: Width of interest.
    :return: KHIVA Array with the result of calculated coefficients.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.cwt_coefficients(ctypes.pointer(tss.arr_reference),
                                                    ctypes.pointer(widths.arr_reference),
                                                    ctypes.pointer(ctypes.c_int(coeff)),
                                                    ctypes.pointer(ctypes.c_int(w)),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def energy_ratio_by_chunks(arr, num_segments, segment_focus):
    """ Calculates the sum of squares of chunk i out of N chunks expressed as a ratio
    with the sum of squares over the whole series. segmentFocus should be lower
    than the number of segments.

    :param arr: KHIVA array with the time series.
    :param num_segments: The number of segments to divide the series into.
    :param segment_focus: The segment number (starting at zero) to return a feature on.
    :return: KHIVA array with the energy ratio by chunk of the time series.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.energy_ratio_by_chunks(ctypes.pointer(arr.arr_reference),
                                                          ctypes.pointer(ctypes.c_long(num_segments)),
                                                          ctypes.pointer(ctypes.c_long(segment_focus)),
                                                          ctypes.pointer(b))

    return Array(array_reference=b)


def fft_aggregated(arr):
    """ Calculates the spectral centroid(mean), variance, skew, and kurtosis of the absolute fourier transform
    spectrum.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the spectral centroid (mean), variance, skew, and kurtosis of the absolute fourier transform
            spectrum.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.fft_aggregated(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def fft_coefficient(arr, coefficient):
    """ Calculates the fourier coefficients of the one-dimensional discrete
    Fourier Transform for real input by fast fourier transformation algorithm.

    :param arr: KHIVA array with the time series.
    :param coefficient: The coefficient to extract from the FFT.
    :return: Tuple with:
        real: KHIVA array with the real part of the coefficient.
        imag: KHIVA array with the imaginary part of the coefficient.
        abs: KHIVA array with the absolute value of the coefficient.
        angle: KHIVA array with the angle of the coefficient.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)
    e = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.fft_coefficient(ctypes.pointer(arr.arr_reference),
                                                   ctypes.pointer(ctypes.c_long(coefficient)),
                                                   ctypes.pointer(b),
                                                   ctypes.pointer(c),
                                                   ctypes.pointer(d),
                                                   ctypes.pointer(e)
                                                   )

    return Array(array_reference=b), Array(array_reference=c,
                                           khiva_type=arr.khiva_type), Array(
        array_reference=d), Array(array_reference=e)


def first_location_of_maximum(arr):
    """ Calculates the first relative location of the maximal value for each time series.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the first relative location of the maximum value to the length of the time series, for each
            time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.first_location_of_maximum(ctypes.pointer(arr.arr_reference),
                                                             ctypes.pointer(b))

    return Array(array_reference=b)


def first_location_of_minimum(arr):
    """ Calculates the first location of the minimal value of each time series. The position is calculated relatively
    to the length of the series.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the first relative location of the minimal value of each series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.first_location_of_minimum(ctypes.pointer(arr.arr_reference),
                                                             ctypes.pointer(b))

    return Array(array_reference=b)


def friedrich_coefficients(arr, m, r):
    """ Coefficients of polynomial :math:`h(x)`, which has been fitted to the deterministic dynamics of Langevin model:
    Largest fixed point of dynamics  :math:`argmax_x {h(x)=0}` estimated from polynomial :math:`h(x)`,
    which has been fitted to the deterministic dynamics of Langevin model:

    .. math::
        \\dot(x)(t) = h(x(t)) + R \\mathcal(N)(0,1)

    as described by [1]. For short time series this method is highly dependent on the parameters.

    [1] Friedrich et al. (2000): Physics Letters A 271, p. 217-222
    Extracting model equations from experimental data.


    :param arr: KHIVA array with the time series.
    :param m: Order of polynom to fit for estimating fixed points of dynamics.
    :param r: Number of quantiles to use for averaging.
    :return: KHIVA array with the coefficients for each time seriess.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.friedrich_coefficients(ctypes.pointer(arr.arr_reference),
                                                          ctypes.pointer(ctypes.c_int(m)),
                                                          ctypes.pointer(ctypes.c_float(r)),
                                                          ctypes.pointer(b))

    return Array(array_reference=b)


def has_duplicates(arr):
    """ Calculates if the input time series contain duplicated elements.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array containing True if the time series contains duplicated elements
     and false otherwise.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.has_duplicates(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def has_duplicate_max(arr):
    """ Calculates if the maximum within input time series is duplicated.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array containing True if the maximum value of the time series is duplicated and false otherwise.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.has_duplicate_max(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def has_duplicate_min(arr):
    """ Calculates if the minimum of the input time series is duplicated.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array containing True if the minimum of the time series is duplicated and False otherwise.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.has_duplicate_min(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def index_mass_quantile(arr, q):
    """ Calculates the index of the mass quantile.

    :param arr: KHIVA array with the time series.
    :param q: The quantile.
    :return: KHIVA array with the index of the mass quantile q.
    """

    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.index_mass_quantile(ctypes.pointer(arr.arr_reference),
                                                       ctypes.pointer(ctypes.c_float(q)),
                                                       ctypes.pointer(b))

    return Array(array_reference=b)


def kurtosis(arr):
    """ Returns the kurtosis of tss (calculated with the adjusted Fisher-Pearson
    standardized moment coefficient G2).

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the kurtosis of each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.kurtosis(ctypes.pointer(arr.arr_reference),
                                            ctypes.pointer(b))

    return Array(array_reference=b)


def large_standard_deviation(arr, r):
    """ Checks if the time series within tss have a large standard deviation.

    :param arr: KHIVA array with the time series.
    :param r: The threshold.
    :return: KHIVA array containing True for those time series in tss that have a large standard deviation.
    """

    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.large_standard_deviation(ctypes.pointer(arr.arr_reference),
                                                            ctypes.pointer(ctypes.c_float(r)),
                                                            ctypes.pointer(b))

    return Array(array_reference=b)


def last_location_of_maximum(arr):
    """ Calculates the last location of the maximum value of each time series. The position
    is calculated relatively to the length of the series.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the last relative location of the maximum value of each series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.last_location_of_maximum(ctypes.pointer(arr.arr_reference),
                                                            ctypes.pointer(b))

    return Array(array_reference=b)


def last_location_of_minimum(arr):
    """ Calculates the last location of the minimum value of each time series. The position
    is calculated relatively to the length of the series.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array the last relative location of the minimum value of each series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.last_location_of_minimum(ctypes.pointer(arr.arr_reference),
                                                            ctypes.pointer(b))

    return Array(array_reference=b)


def length(arr):
    """ Returns the length of the input time series.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array the length of tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.length(ctypes.pointer(arr.arr_reference),
                                          ctypes.pointer(b))

    return Array(array_reference=b)


def linear_trend(arr):
    """ Calculate a linear least-squares regression for the values of the time series versus the sequence from 0 to
    length of the time series minus one.

    :param arr: KHIVA array with the time series.
    :return  a tuple with:
            pvalue: KHIVA array the pvalues for all time series.
            rvalue: KHIVA array The rvalues for all time series.
            intercept: KHIVA array the intercept values for all time series.
            slope: KHIVA array the slope for all time series.
            stdrr: KHIVA array the stderr values for all time series.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)
    e = ctypes.c_void_p(0)
    f = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.linear_trend(ctypes.pointer(arr.arr_reference),
                                                ctypes.pointer(b),
                                                ctypes.pointer(c),
                                                ctypes.pointer(d),
                                                ctypes.pointer(e),
                                                ctypes.pointer(f)
                                                )

    return Array(array_reference=b), \
           Array(array_reference=c), \
           Array(array_reference=d), \
           Array(array_reference=e), \
           Array(array_reference=f)


def local_maximals(arr):
    """ Calculates all Local Maximals fot the time series in array.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the calculated local maximals for each time series in arr.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.local_maximals(ctypes.pointer(arr.arr_reference),
                                                  ctypes.pointer(b))

    return Array(array_reference=b)


def longest_strike_above_mean(arr):
    """ Calculates the length of the longest consecutive subsequence in tss that is bigger than the mean of tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the length of the longest consecutive subsequence in the input time series that is bigger
            than the mean.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.longest_strike_above_mean(ctypes.pointer(arr.arr_reference),
                                                             ctypes.pointer(b))

    return Array(array_reference=b)


def longest_strike_below_mean(arr):
    """ Calculates the length of the longest consecutive subsequence in tss that is below the mean of tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the length of the longest consecutive subsequence in the input time series that is below
            the mean.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.longest_strike_below_mean(ctypes.pointer(arr.arr_reference),
                                                             ctypes.pointer(b))

    return Array(array_reference=b)


def max_langevin_fixed_point(arr, m, r):
    """ Largest fixed point of dynamics  :math:`argmax_x {h(x)=0}` estimated from polynomial :math:`h(x)`,
    which has been fitted to the deterministic dynamics of Langevin model

    .. math::
        \\dot(x)(t) = h(x(t)) + R \\mathcal(N)(0,1)

    as described by

        Friedrich et al. (2000): Physics Letters A 271, p. 217-222
        *Extracting model equations from experimental data*

    :param arr: KHIVA array with the time series.
    :param m: Order of polynom to fit for estimating fixed points of dynamics.
    :param r: Number of quantiles to use for averaging.
    :return: KHIVA array with the largest fixed point of deterministic dynamics.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.max_langevin_fixed_point(ctypes.pointer(arr.arr_reference),
                                                            ctypes.pointer(ctypes.c_int(m)),
                                                            ctypes.pointer(ctypes.c_float(r)),
                                                            ctypes.pointer(b))

    return Array(array_reference=b)


def maximum(arr):
    """ Calculates the maximum value for each time series within tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the maximum value of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.maximum(ctypes.pointer(arr.arr_reference),
                                           ctypes.pointer(b))

    return Array(array_reference=b)


def mean(arr):
    """ Calculates the mean value for each time series within tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the mean value of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.mean(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def mean_absolute_change(arr):
    """ Calculates the mean over the absolute differences between subsequent time series values in tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the mean over the absolute differences between subsequent time series values.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.mean_absolute_change(ctypes.pointer(arr.arr_reference),
                                                        ctypes.pointer(b))

    return Array(array_reference=b)


def mean_change(arr):
    """ Calculates the mean over the differences between subsequent time series values in tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the mean over the differences between subsequent time series values.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.mean_change(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def mean_second_derivative_central(arr):
    """ Calculates mean value of a central approximation of the second derivative for each time series in tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the mean value of a central approximation of the second derivative for each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.mean_second_derivative_central(ctypes.pointer(arr.arr_reference),
                                                                  ctypes.pointer(b))

    return Array(array_reference=b)


def median(arr):
    """ Calculates the median value for each time series within tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the median value of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.median(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def minimum(arr):
    """ Calculates the minimum value for each time series within tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the minimum value of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.minimum(ctypes.pointer(arr.arr_reference),
                                           ctypes.pointer(b))

    return Array(array_reference=b)


def number_crossing_m(arr, m):
    """ Calculates the number of m-crossings. A m-crossing is defined as two sequential values where the first
    value is lower than m and the next is greater, or viceversa. If you set m to zero, you will get the number of
    zero crossings.

    :param arr: KHIVA array with the time series.
    :param m: The m value.
    :return: KHIVA array with the number of m-crossings of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.number_crossing_m(ctypes.pointer(arr.arr_reference),
                                                     ctypes.pointer(ctypes.c_int(m)),
                                                     ctypes.pointer(b))

    return Array(array_reference=b)


def number_cwt_peaks(arr, max_w):
    """ This feature calculator searches for different peaks. To do so, the time series is smoothed by a ricker
    wavelet and for widths ranging from 1 to :math:'max_w`. This feature calculator returns the number of peaks that
    occur at enough width scales and with sufficiently high Signal-to-Noise-Ratio (SNR).

    :param arr: KHIVA array with the time series.
    :param max_w: The maximum width to consider.
    :return: KHIVA array with the number of peaks for each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.number_cwt_peaks(ctypes.pointer(arr.arr_reference),
                                                    ctypes.pointer(ctypes.c_int(max_w)),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def number_peaks(arr, n):
    """ Calculates the number of peaks of at least support :math:`n` in the time series :math:`tss`. A peak of support
    :math:`n` is defined as a subsequence of :math:`tss where a value occurs, which is bigger than
    its :math:`n` neighbours to the left and to the right.

    :param arr: KHIVA array with the time series.
    :param n: The support of the peak.
    :return: KHIVA array with the number of peaks of at least support :math:`n`.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.number_peaks(ctypes.pointer(arr.arr_reference), ctypes.pointer(ctypes.c_int(n)),
                                                ctypes.pointer(b))

    return Array(array_reference=b)


def partial_autocorrelation(arr, lags):
    """ Calculates the value of the partial autocorrelation function at the given lag. The lag :math:`k'`  partial
    autocorrelation of a time series :math:`\\lbrace x_t, t = 1 \\ldots T \\rbrace` equals the partial correlation of
    :math:`x_t` and :math:`x_{t-k}`, adjusted for the intermediate variables :math:`\\lbrace x_{t-1}, \\ldots, x_{t-k+1}\\rbrace`
    ([1]). Following [2], it can be defined as:

    .. math::
          \\alpha_k = \\frac{ Cov(x_t, x_{t-k} | x_{t-1}, \\ldots, x_{t-k+1})}
          {\\sqrt{ Var(x_t | x_{t-1}, \\ldots, x_{t-k+1}) Var(x_{t-k} | x_{t-1}, \\ldots, x_{t-k+1} )}}

    with (a) :math:`x_t = f(x_{t-1}, \\ldots, x_{t-k+1})` and (b) :math:`x_{t-k} = f(x_{t-1}, \\ldots, x_{t-k+1})`
    being AR(k-1) models that can be fitted by OLS. Be aware that in (a), the regression is done on past values to
    predict :math:`x_t` whereas in (b), future values are used to calculate the past value :math:`x_{t-k}`.
    It is said in [1] that "for an AR(p), the partial autocorrelations :math:`\\alpha_k` will be nonzero for :math:`k<=p`
    and zero for :math:`k>p`."
    With this property, it is used to determine the lag of an AR-Process.

    [1] Box, G. E., Jenkins, G. M., Reinsel, G. C., & Ljung, G. M. (2015).
    Time series analysis: forecasting and control. John Wiley & Sons.
    [2] https://onlinecourses.science.psu.edu/stat510/node/62

    :param arr: KHIVA array with the time series.
    :param lags: KHIVA array with the lags to be calculated.
    :return: KHIVA array with the partial autocorrelation for each time series for the given lag.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.partial_autocorrelation(ctypes.pointer(arr.arr_reference),
                                                           ctypes.pointer(lags.arr_reference),
                                                           ctypes.pointer(b))

    return Array(array_reference=b)


def percentage_of_reoccurring_datapoints_to_all_datapoints(arr, is_sorted):
    """ Calculates the percentage of unique values, that are present in the time series more than once.

    .. math::

        len(different values occurring more than once) / len(different values)

    This means the percentage is normalized to the number of unique values, in contrast to the
    percentage_of_reoccurring_values_to_all_values.

    :param arr: KHIVA array with the time series.
    :param is_sorted: Indicates if the input time series is sorted or not. Defaults to false.
    :return: KHIVA array with the percentage of unique values, that are present in the time series more than once.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.percentage_of_reoccurring_datapoints_to_all_datapoints(
        ctypes.pointer(arr.arr_reference),
        ctypes.pointer(
            ctypes.c_bool(is_sorted)),
        ctypes.pointer(
            b))

    return Array(array_reference=b)


def percentage_of_reoccurring_values_to_all_values(arr, is_sorted):
    """ Calculates the percentage of unique values, that are present in the time series more than once.

    .. math::

        \\frac{\\textit{number of data points occurring more than once}}{\\textit{number of all data points})}

    This means the percentage is normalized to the number of unique values, in contrast to the
    percentage_of_reoccurring_datapoints_to_all_datapoints.

    :param arr: KHIVA array with the time series.
    :param is_sorted: Indicates if the input time series is sorted or not. Defaults to false.
    :return: KHIVA array with the percentage of unique values, that are present in the time series more than once.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.percentage_of_reoccurring_values_to_all_values(ctypes.pointer(arr.arr_reference),
                                                                                  ctypes.pointer(
                                                                                      ctypes.c_bool(is_sorted)),
                                                                                  ctypes.pointer(b))

    return Array(array_reference=b)


def quantile(arr, q, precision=1e8):
    """ Returns values at the given quantile.

    :param arr: KHIVA array with the time series.
    :param q: Khiva array with the percentile(s) at which to extract score(s). One or many.
    :param precision: Number of decimals expected.
    :return: Values at the given quantile.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.quantile(ctypes.pointer(arr.arr_reference), ctypes.pointer(q.arr_reference),
                                            ctypes.pointer(ctypes.c_float(precision)),
                                            ctypes.pointer(b))

    return Array(array_reference=b)


def range_count(arr, min, max):
    """ Counts observed values within the interval [min, max).

    :param arr: KHIVA array with the time series.
    :param min: Value that sets the lower limit.
    :param max: Value that sets the upper limit.
    :return: KHIVA array with the values at the given range.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.range_count(ctypes.pointer(arr.arr_reference),
                                               ctypes.pointer(ctypes.c_int(min)),
                                               ctypes.pointer(ctypes.c_float(max)),
                                               ctypes.pointer(b))

    return Array(array_reference=b)


def ratio_beyond_r_sigma(arr, r):
    """ Calculates the ratio of values that are more than :math:`r*std(x)` (so :math:`r` sigma) away from the mean of
    :math:`x`.

    :param arr: KHIVA array with the time series.
    :param r: Number of times that the values should be away from.
    :return: KHIVA array with the ratio of values that are more than :math:`r*std(x)` (so :math:`r` sigma) away from
            the mean of :math:`x`.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.ratio_beyond_r_sigma(ctypes.pointer(arr.arr_reference),
                                                        ctypes.pointer(ctypes.c_float(r)),
                                                        ctypes.pointer(b))

    return Array(array_reference=b)


def ratio_value_number_to_time_series_length(arr):
    """ Calculates a factor which is 1 if all values in the time series occur only once, and below one if this is
    not the case. In principle, it just returns:

    .. math::
        \\frac{\\textit{number unique values}}{\\textit{number values}}

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the ratio of unique values with respect to the total number of values.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.ratio_value_number_to_time_series_length(ctypes.pointer(arr.arr_reference),
                                                                            ctypes.pointer(b))

    return Array(array_reference=b)


def sample_entropy(arr):
    """ Calculates a vectorized sample entropy algorithm.
    https://en.wikipedia.org/wiki/Sample_entropy
    https://www.ncbi.nlm.nih.gov/pubmed/10843903?dopt=Abstract
    For short time-series this method is highly dependent on the parameters, but should be stable for N > 2000,
    see: Yentes et al. (2012) - The Appropriate Use of Approximate Entropy and Sample Entropy with Short Data Sets
    Other shortcomings and alternatives discussed in:
    Richman & Moorman (2000) - Physiological time-series analysis using approximate entropy and sample entropy.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the same dimensions as tss, whose values (time series in dimension 0)
            contains the vectorized sample entropy for all the input time series in tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.sample_entropy(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def skewness(arr):
    """ Calculates the sample skewness of tss (calculated with the adjusted Fisher-Pearson standardized
    moment coefficient G1).

    :param arr: KHIVA array with the time series.
    :return: KHIVA array containing the skewness of each time series in tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.skewness(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def spkt_welch_density(arr, coeff):
    """ Estimates the cross power spectral density of the time series array at different frequencies. To do so, the
    time series is first shifted from the time domain to the frequency domain.

    Welch's method computes an estimate of the power spectral density by dividing the data into overlapping
    segments, computing a modified periodogram for each segment and averaging the periodograms.
    [1] P. Welch, "The use of the fast Fourier transform for the estimation of power spectra: A method based on time
    averaging over short, modified periodograms", IEEE Trans. Audio Electroacoust. vol. 15, pp. 70-73, 1967.
    [2] M.S. Bartlett, "Periodogram Analysis and Continuous Spectra", Biometrika, vol. 37, pp. 1-16, 1950.
    [3] Rabiner, Lawrence R., and B. Gold. "Theory and Application of Digital Signal Processing" Prentice-Hall, pp.
    414-419, 1975.

    :param arr: KHIVA array with the time series.
    :param coeff: The coefficient to be returned.
    :return: KHIVA array containing the power spectrum of the different frequencies for each time series in arr.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.spkt_welch_density(ctypes.pointer(arr.arr_reference),
                                                      ctypes.pointer(ctypes.c_int(coeff)), ctypes.pointer(b))

    return Array(array_reference=b)


def standard_deviation(arr):
    """ Calculates the standard deviation of each time series within tss.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array with the standard deviation of each time series within tss.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.standard_deviation(ctypes.pointer(arr.arr_reference),
                                                      ctypes.pointer(b))

    return Array(array_reference=b)


def sum_of_reoccurring_datapoints(arr, is_sorted=False):
    """ Calculates the sum of all data points, that are present in the time series more than once.

    :param arr: KHIVA array with the time series.
    :param is_sorted: Indicates if the input time series is sorted or not. Defaults to false.
    :return: KHIVA array with the sum of all data points, that are present in the time series more than once.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.sum_of_reoccurring_datapoints(ctypes.pointer(arr.arr_reference),
                                                                 ctypes.pointer(ctypes.c_bool(is_sorted)),
                                                                 ctypes.pointer(b))

    return Array(array_reference=b)


def sum_of_reoccurring_values(arr, is_sorted=False):
    """ Calculates the sum of all values, that are present in the time series more than once.

    :param arr: KHIVA array with the time series.
    :param is_sorted: Indicates if the input time series is sorted or not. Defaults to false.
    :return: KHIVA array with the sum of all values, that are present in the time series more than once.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.sum_of_reoccurring_values(ctypes.pointer(arr.arr_reference),
                                                             ctypes.pointer(ctypes.c_bool(is_sorted)),
                                                             ctypes.pointer(b))

    return Array(array_reference=b)


def sum_values(arr):
    """ Calculates the sum over the time series arr.

    :param arr: KHIVA array with the time series.
    :param is_sorted: Indicates if the input time series is sorted or not. Defaults to false.
    :return: KHIVA array with the sum of values in each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.sum_values(ctypes.pointer(arr.arr_reference),
                                              ctypes.pointer(b))

    return Array(array_reference=b)


def symmetry_looking(arr, r):
    """ Calculates if the distribution of tss *looks symmetric*. This is the case if

    .. math::

         | mean(tss)-median(tss)| < r * (max(tss)-min(tss))


    :param arr: KHIVA array with the time series.
    :param r: The percentage of the range to compare with.
    :return: KHIVA array denoting if the input time series look symmetric.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.symmetry_looking(ctypes.pointer(arr.arr_reference),
                                                    ctypes.pointer(ctypes.c_float(r)),
                                                    ctypes.pointer(b))

    return Array(array_reference=b)


def time_reversal_asymmetry_statistic(arr, lag):
    """ This function calculates the value of:

    .. math::

        \\frac{1}{n-2lag} \\sum_{i=0}^{n-2lag} x_{i + 2 \\cdot lag}^2 \\cdot x_{i + lag} - x_{i + lag} \\cdot  x_{i}^2

    which is

    .. math::

        \\mathbb{E}[L^2(X)^2 \\cdot L(X) - L(X) \\cdot X^2]

    where :math:`\\mathbb{E}` is the mean and :math:`L` is the lag operator. It was proposed in [1] as a promising
    feature to extract from time series.

    :param arr: KHIVA array with the time series.
    :param lag: The lag to be computed.
    :return: KHIVA array containing the count of the given value in each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.time_reversal_asymmetry_statistic(ctypes.pointer(arr.arr_reference),
                                                                     ctypes.pointer(ctypes.c_int(lag)),
                                                                     ctypes.pointer(b))

    return Array(array_reference=b)


def value_count(arr, v):
    """ Counts occurrences of value in the time series tss.

    :param arr: KHIVA array with the time series.
    :param v: The value to be counted.
    :return: KHIVA array containing the count of the given value in each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.value_count(ctypes.pointer(arr.arr_reference), ctypes.pointer(ctypes.c_float(v)),
                                               ctypes.pointer(b))

    return Array(array_reference=b)


def variance(arr):
    """ Computes the variance for the time series array.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array containing the variance in each time series.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.variance(ctypes.pointer(arr.arr_reference), ctypes.pointer(b))

    return Array(array_reference=b)


def variance_larger_than_standard_deviation(arr):
    """ Calculates if the variance of array is greater than the standard deviation. In other words, if the variance of
    array is larger than 1.

    :param arr: KHIVA array with the time series.
    :return: KHIVA array denoting if the variance of array is greater than the standard deviation.
    """
    b = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.variance_larger_than_standard_deviation(ctypes.pointer(arr.arr_reference),
                                                                           ctypes.pointer(b))

    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################


def linear(xss, yss):
    """Calculates a linear least-squares regression for two sets of measurements. Both arrays should have the same
    length.

    :param xss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.
    :param yss: Expects an input array whose dimension zero is the length of the time series (all the same) and dimension
                one indicates the number of time series.

    :return: slope Slope of the regression line.
            intercept Intercept of the regression line.
            rvalue Correlation coefficient.
            pvalue Two-sided p-value for a hypothesis test whose null hypothesis is that the slope is zero, using Wald
            Test with t-distribution of the test statistic.
            stderrest Standard error of the estimated gradient.
    """
    b = ctypes.c_void_p(0)
    c = ctypes.c_void_p(0)
    d = ctypes.c_void_p(0)
    e = ctypes.c_void_p(0)
    f = ctypes.c_void_p(0)

    KhivaLibrary().c_khiva_library.linear(ctypes.pointer(xss.arr_reference),
                                          ctypes.pointer(yss.arr_reference),
                                          ctypes.pointer(b),
                                          ctypes.pointer(c),
                                          ctypes.pointer(d),
                                          ctypes.pointer(e),
                                          ctypes.pointer(f)
                                          )
    return Array(array_reference=b), Array(array_reference=c), Array(array_reference=d), Array(
        array_reference=e), Array(array_reference=f)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################

import ctypes
import logging
import sys
from enum import Enum

import numpy as np
import pandas as pd

from khiva.library import KhivaLibrary


########################################################################################################################

class dtype(Enum):
    """
    KHIVA array available types.
    """
    f32 = 0
    """
    Float. khiva.dtype
    """
    c32 = 1
    """
    32 bits Complex. khiva.dtype
    """
    f64 = 2
    """
    64 bits Double. khiva.dtype
    """
    c64 = 3
    """
    64 bits Complex. khiva.dtype
    """
    b8 = 4
    """
    Boolean. khiva.dtype
    """
    s32 = 5
    """
    32 bits Int. khiva.dtype
    """
    u32 = 6
    """
    32 bits Unsigned Int. khiva.dtype
    """
    u8 = 7
    """
    8 bits Unsigned Int. khiva.dtype
    """
    s64 = 8
    """
    64 bits Integer. khiva.dtype
    """
    u64 = 9
    """
    64 bits Unsigned Int. khiva.dtype
    """
    s16 = 10
    """
    16 bits Int. khiva.dtype
    """
    u16 = 11
    """
    16 bits Unsigned int. khiva.dtype
    """


def _get_array_type(khiva_type):
    """
    Transform the KHIVA type to its equivalent in ctypes.

    :param khiva_type: KHIVA type.

    :return: The ctypes equivalent.
    """
    return {
        dtype.f32.value: ctypes.c_float,
        dtype.c32.value: ctypes.c_float,
        dtype.f64.value: ctypes.c_double,
        dtype.c64.value: ctypes.c_double,
        dtype.b8.value: ctypes.c_bool,
        dtype.u8.value: ctypes.c_uint8,
        dtype.s16.value: ctypes.c_int16,
        dtype.u16.value: ctypes.c_uint16,
        dtype.s32.value: ctypes.c_int32,
        dtype.u32.value: ctypes.c_uint32,
        dtype.s64.value: ctypes.c_int64,
        dtype.u64.value: ctypes.c_uint64
    }[khiva_type]


def _get_numpy_type(khiva_type):
    """
     Transform the KHIVA type to its equivalent in Numpy.

    :param khiva_type: KHIVA type.

    :return: The Numpy type equivalent.
    """
    return {
        dtype.f32.value: np.float,
        dtype.c32.value: np.complex64,
        dtype.f64.value: np.double,
        dtype.c64.value: np.complex128,
        dtype.b8.value: np.bool,
        dtype.u8.value: np.uint8,
        dtype.s16.value: np.int8,
        dtype.u16.value: np.uint16,
        dtype.s32.value: np.int16,
        dtype.u32.value: np.uint32,
        dtype.s64.value: np.int64,
        dtype.u64.value: np.uint64,
    }[khiva_type]


class Array:
    __array_priority__ = 50

    def __init__(self, data=None, khiva_type=dtype.f32, array_reference=None, arrayfire_reference=False):
        """
        Creates a KHIVA array in one of the following ways: 1) using a previously created array; or 2) with data (in
        numpy, list, or pandas dataframe format)

        :param data: Numpy array, List of elements or a Pandas dataframe.
        :param khiva_type: KHIVA type.
        :param array_reference: Reference of the array.
        """
        if array_reference is None:
            self.khiva_type = khiva_type
            self.arr_reference = self._create_array(data)
            self.dims = self.get_dims()
            self.result_l = self._get_result_length()
        else:
            self.arr_reference = array_reference
            self.khiva_type = self.get_type()
            self.dims = self.get_dims()
            self.result_l = self._get_result_length()

        self.arrayfire_reference = arrayfire_reference

    @classmethod
    def from_arrayfire(cls, arrayfire):
        """
        Creates a KHIVA array from an array of ArrayFire.

        :param arrayfire: An ArrayFire array.
        :return: a KHIVA array.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.from_arrayfire(ctypes.pointer(arrayfire.arr), ctypes.pointer(result))
        arrayfire.arr.value = 0
        return cls(array_reference=result, arrayfire_reference=False)

    def _create_array(self, data):
        """ Creates the KHIVA array in the device.

        :param data: The data used for creating the khiva array.

        :return An opaque pointer to the Array.
        """
        if isinstance(data, list):
            data = np.asarray(data)
        if isinstance(data, pd.DataFrame):
            data = data.values
        shape = np.array(data.shape)

        if data.size > 1:
            trimmed_dims = shape
            for _ in range(0, 3):
                if trimmed_dims[-1] == 1:
                    trimmed_dims = trimmed_dims[:-1]
            shape = trimmed_dims[::-1]
        else:
            shape = np.array([1])

        c_array_n = (ctypes.c_longlong * len(shape))(*(np.array(shape)).astype(np.longlong))
        c_ndims = ctypes.c_uint(len(shape))
        c_complex = np.iscomplexobj(data)

        if c_complex:
            data = np.dstack((data.real.flatten(), data.imag.flatten()))

        array_joint = data.flatten()

        c_array_joint = (_get_array_type(self.khiva_type.value) * len(array_joint))(
            *array_joint)
        opaque_pointer = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.create_array(ctypes.pointer(c_array_joint),
                                                    ctypes.pointer(c_ndims),
                                                    ctypes.pointer(c_array_n),
                                                    ctypes.pointer(opaque_pointer),
                                                    ctypes.pointer(ctypes.c_int(self.khiva_type.value)))
        return opaque_pointer

    def _get_data(self):
        """ Retrieves the data from the device to the host.

        :return A numpy array with the data.
        """
        initialized_result_array = np.zeros(self.result_l).astype(_get_array_type(self.khiva_type.value))
        c_result_array = (_get_array_type(self.khiva_type.value) * self.result_l)(*initialized_result_array)
        KhivaLibrary().c_khiva_library.get_data(ctypes.pointer(self.arr_reference), ctypes.pointer(c_result_array))

        a = np.array(c_result_array)

        if self._is_complex():
            a = np.array(np.split(a, self.result_l / 2))
            a = np.apply_along_axis(lambda args: [complex(*args)], 1, a)

        # Clean up the last n dimensions if these are equal to 1
        if a.size > 1:
            trimmed_dims = self.get_dims()
            for _ in range(0, 3):
                if trimmed_dims[-1] == 1:
                    trimmed_dims = trimmed_dims[:-1]
        else:
            trimmed_dims = np.array([1])

        a = a.reshape(trimmed_dims[::-1])

        a = a.astype(_get_numpy_type(self.khiva_type.value))
        return a

    def _get_result_length(self):
        """ Gets the length of the result.

        :return: The length of the result, used in order to get the data to the host.
        """
        result = 1
        for value in self.dims:
            result *= value

        if self._is_complex():
            result *= 2

        return int(result)

    def get_dims(self):
        """ Gets the dimensions of the KHIVA array.

        :return: The dimensions of the KHIVA array.
        """
        c_array_n = (ctypes.c_longlong * 4)(*(np.zeros(4)).astype(np.longlong))
        KhivaLibrary().c_khiva_library.get_dims(ctypes.pointer(self.arr_reference), ctypes.pointer(c_array_n))
        return np.array(c_array_n)

    def get_type(self):
        """ Gets the type of the KHIVA array.

        :return: The type of the KHIVA array.
        """
        c_type = ctypes.c_int()
        KhivaLibrary().c_khiva_library.get_type(ctypes.pointer(self.arr_reference), ctypes.pointer(c_type))
        return dtype(c_type.value)

    def _is_complex(self):
        """ Returns True if the array contains complex numbers and False otherwise.

        :return: True if the array contains complex numbers and False otherwise.
        """
        return self.khiva_type.value == dtype.c32.value or self.khiva_type.value == dtype.c64.value

    def to_arrayfire(self):
        """ Creates an Arrayfire array from this KHIVA array. This need to be used carefully as the same array
        reference is oging to be used by both of them. Once the Arrayfire array is created, the destructor of
        the KHIVA array is not going to free the allocated array.

        :return: an Arrayfire Array
        """
        try:
            import arrayfire as af
        except ModuleNotFoundError:
            logging.error("In order to use `to_arrayfire()` function, you need to install the Arrayfire Python library")
            sys.exit(1)
        result = af.Array()
        result.arr = self.arr_reference
        self.arrayfire_reference = True
        return result

    def to_list(self):
        """ Converts the KHIVA array to a list.

        :return: KHIVA array converted to list.
        """
        return self._get_data().tolist()

    def to_numpy(self):
        """ Converts the KHIVA array to a numpy array.

        The returned numpy array shape matches the Array dimensions as follows:
          - For an Array with dims equal to [4, 2, 1, 1] the numpy shape will be (2, 4).
          - For an Array with dims equal to [4, 3, 2, 1] the numpy shape will be (2, 3, 4).
          - For an Array with dims equal to [4, 1, 2, 3] the numpy shape will be (3, 2, 1, 4).

        :return: KHIVA array converted to numpy.array.
        """
        return self._get_data()

    def to_pandas(self):
        """ Converts the KHIVA array to a pandas data frame.

        :return: KHIVA array converted to a pandas data frame.
        """
        return pd.DataFrame(data=self._get_data())

    def display(self):
        """
        Displays the data stored in the KHIVA array.
        """
        KhivaLibrary().c_khiva_library.display(ctypes.pointer(self.arr_reference))

    def join(self, dim, other):
        """
        Joins the first and second KHIVA arrays along the specified dimension.
        :param dim: The dimension along which the join occurs.
        :param other: The second input array.
        :return: KHIVA Array with the result of this operation.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.join(ctypes.pointer(ctypes.c_int(dim)),
                                            ctypes.pointer(self.arr_reference),
                                            ctypes.pointer(other.arr_reference),
                                            ctypes.pointer(result))
        return Array(array_reference=result)

    def __len__(self):
        """
        Return the length.
        """
        if self._is_complex():
            return self.result_l / 2
        else:
            return self.result_l

    def __del__(self):
        """
        Class destructor.
        """
        if not self.arrayfire_reference:
            KhivaLibrary().c_khiva_library.delete_array(ctypes.pointer(self.arr_reference))

    def __add__(self, other):
        """
        Return self + other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_add(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __iadd__(self, other):
        """
        Perform self += other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_add(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __radd__(self, other):
        """
        Return other + self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_add(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __sub__(self, other):
        """
        Return self - other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_sub(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __isub__(self, other):
        """
        Perform self -= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_sub(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rsub__(self, other):
        """
        Return other - self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_sub(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __mul__(self, other):
        """
        Return self * other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mul(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __imul__(self, other):
        """
        Perform self *= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mul(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rmul__(self, other):
        """
        Return other * self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mul(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __truediv__(self, other):
        """
        Return self / other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __itruediv__(self, other):
        """
        Perform self /= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rtruediv__(self, other):
        """
        Return other / self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __div__(self, other):
        """
        Return self / other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __idiv__(self, other):
        """
        Perform other / self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rdiv__(self, other):
        """
        Return other / self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_div(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __mod__(self, other):
        """
        Return self % other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mod(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __imod__(self, other):
        """
        Perform self %= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mod(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rmod__(self, other):
        """
        Return other % self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_mod(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __pow__(self, other):
        """
        Return self ** other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_pow(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __ipow__(self, other):
        """
        Perform self **= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_pow(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __rpow__(self, other):
        """
        Return other ** self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_pow(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(other.arr_reference),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def __lt__(self, other):
        """
        Return self < other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_lt(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __gt__(self, other):
        """
        Return self > other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_gt(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __le__(self, other):
        """
        Return self <= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_le(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __ge__(self, other):
        """
        Return self >= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_ge(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __eq__(self, other):
        """
        Return self == other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_eq(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __ne__(self, other):
        """
        Return self != other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_ne(ctypes.pointer(self.arr_reference), ctypes.pointer(other.arr_reference),
                                                ctypes.pointer(result))
        return Array(array_reference=result)

    def __and__(self, other):
        """
        Return self & other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitand(ctypes.pointer(self.arr_reference),
                                                    ctypes.pointer(other.arr_reference),
                                                    ctypes.pointer(result))
        return Array(array_reference=result)

    def __iand__(self, other):
        """
        Perform self &= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitand(ctypes.pointer(self.arr_reference),
                                                    ctypes.pointer(other.arr_reference),
                                                    ctypes.pointer(result))
        return Array(array_reference=result)

    def __or__(self, other):
        """
        Return self | other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitor(ctypes.pointer(self.arr_reference),
                                                   ctypes.pointer(other.arr_reference),
                                                   ctypes.pointer(result))
        return Array(array_reference=result)

    def __ior__(self, other):
        """
        Perform self |= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitor(ctypes.pointer(self.arr_reference),
                                                   ctypes.pointer(other.arr_reference),
                                                   ctypes.pointer(result))
        return Array(array_reference=result)

    def __xor__(self, other):
        """
        Return self ^ other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitxor(ctypes.pointer(self.arr_reference),
                                                    ctypes.pointer(other.arr_reference),
                                                    ctypes.pointer(result))
        return Array(array_reference=result)

    def __ixor__(self, other):
        """
        Perform self ^= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitxor(ctypes.pointer(self.arr_reference),
                                                    ctypes.pointer(other.arr_reference),
                                                    ctypes.pointer(result))
        return Array(array_reference=result)

    def __lshift__(self, other):
        """
        Return self << other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitshiftl(ctypes.pointer(self.arr_reference),
                                                       ctypes.pointer(ctypes.c_int32(other)), ctypes.pointer(result))
        return Array(array_reference=result)

    def __ilshift__(self, other):
        """
        Perform self <<= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitshiftl(ctypes.pointer(self.arr_reference),
                                                       ctypes.pointer(ctypes.c_int32(other)), ctypes.pointer(result))
        return Array(array_reference=result)

    def __rshift__(self, other):
        """
        Return self >> other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitshiftr(ctypes.pointer(self.arr_reference),
                                                       ctypes.pointer(ctypes.c_int32(other)), ctypes.pointer(result))
        return Array(array_reference=result)

    def __irshift__(self, other):
        """
        Perform self >>= other.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_bitshiftr(ctypes.pointer(self.arr_reference),
                                                       ctypes.pointer(ctypes.c_int32(other)), ctypes.pointer(result))
        return Array(array_reference=result)

    def __neg__(self):
        """
        Return -self
        """
        return Array(np.zeros(self.get_dims())) - self

    def __pos__(self):
        """
        Return +self
        """
        return self

    def __invert__(self):
        """
        Return ~self
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_not(ctypes.pointer(self.arr_reference), ctypes.pointer(result))
        return Array(array_reference=result)

    def _get_metadata_str(self, dims=True):
        return 'khiva.Array()\nType: {}\n{}' \
            .format(self.khiva_type, 'Dims: {}'.format(str(self.dims)) if dims else '')

    def __str__(self):
        """
        Converts the khiva array to string showing its meta data and contents.
        """

        return self._get_metadata_str()

    def __nonzero__(self):
        """
        Returns if the Array is non-zero.
        """
        ne = self != Array(np.zeros(self.get_dims()))
        ne_host = ne.to_numpy()
        return int(np.all(ne_host))

    def __repr__(self):
        """
        Displays the meta data of the arrayfire array.
        """

        return self._get_metadata_str()

    def transpose(self, conjugate=False):
        """
        Transpose the KHIVA Array.

        :param conjugate: Indicates if the transpose is conjugated or not.
        :return: The transposed KHIVA Array.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_transpose(ctypes.pointer(self.arr_reference),
                                                       ctypes.pointer(ctypes.c_bool(conjugate)), ctypes.pointer(result))

        return Array(array_reference=result)

    def get_col(self, index):
        """
        Gets a desired column.

        :param index: Index of the desired column.
        :return: The desired column.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_col(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(ctypes.c_int32(index)),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def get_cols(self, first, last):
        """
        Gets a sequence of columns using the first column index and the last column index, both columns included.

        :param first: First column of the subsequence of columns.
        :param last: Last column of the subsequence of columns.
        :return: A subsequence of columns between 'first' and 'last'.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_cols(ctypes.pointer(self.arr_reference),
                                                  ctypes.pointer(ctypes.c_int32(first)),
                                                  ctypes.pointer(ctypes.c_int32(last)),
                                                  ctypes.pointer(result))
        return Array(array_reference=result)

    def get_row(self, index):
        """
        Gets a desired row.

        :param index: Index of the desired row.
        :return: The desired row.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_row(ctypes.pointer(self.arr_reference),
                                                 ctypes.pointer(ctypes.c_int32(index)),
                                                 ctypes.pointer(result))
        return Array(array_reference=result)

    def get_rows(self, first, last):
        """
        Gets a sequence of rows using the first row index and the last row index, both rows included.

        :param first: First row of the subsequence of rows.
        :param last: Last row of the subsequence of rows.
        :return: A subsequence of rows between 'first' and 'last'.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_rows(ctypes.pointer(self.arr_reference),
                                                  ctypes.pointer(ctypes.c_int32(first)),
                                                  ctypes.pointer(ctypes.c_int32(last)),
                                                  ctypes.pointer(result))
        return Array(array_reference=result)

    def matmul(self, other):
        """
        Matrix multiplication.

        :param other: KHIVA Array
        :return: The matrix multiplication between these two KHIVA Arrays.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_matmul(ctypes.pointer(self.arr_reference),
                                                    ctypes.pointer(other.arr_reference),
                                                    ctypes.pointer(result))
        return Array(array_reference=result)

    def copy(self):
        """
        Performs a deep copy of the array.

        return: An identical copy of self.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.copy(ctypes.pointer(self.arr_reference), ctypes.pointer(result))
        return Array(array_reference=result)

    def as_type(self, dtype):
        """
        Converts the array to a desired array with a desired type.

        :param dtype: The desired KHIVA data type.
        :return: An array with the desired data type.
        """
        result = ctypes.c_void_p(0)
        KhivaLibrary().c_khiva_library.khiva_as(ctypes.pointer(self.arr_reference),
                                                ctypes.pointer(ctypes.c_int32(dtype.value)), ctypes.pointer(result))
        self.khiva_type = self.get_type()
        return Array(array_reference=result)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import re
import os
import requests
from khiva.library import *

########################################################################################################################

class LibraryTest(unittest.TestCase):

    def setUp(self):
        pass

    def test_get_backend_info(self):
        info = get_backend_info()
        word = info.split()[0]
        self.assertEqual(word, "ArrayFire")

    def test_set_backend(self):
        backends = get_backends()
        cuda = backends & KHIVABackend.KHIVA_BACKEND_CUDA.value
        opencl = backends & KHIVABackend.KHIVA_BACKEND_OPENCL.value
        cpu = backends & KHIVABackend.KHIVA_BACKEND_CPU.value
        b = get_backend()
        if cuda:
            set_backend(KHIVABackend.KHIVA_BACKEND_CUDA)
            self.assertEqual(get_backend(), KHIVABackend.KHIVA_BACKEND_CUDA)
            set_backend(b)

        if opencl:
            set_backend(KHIVABackend.KHIVA_BACKEND_OPENCL)
            self.assertEqual(get_backend(), KHIVABackend.KHIVA_BACKEND_OPENCL)
            set_backend(b)

        if cpu:
            set_backend(KHIVABackend.KHIVA_BACKEND_CPU)
            self.assertEqual(get_backend(), KHIVABackend.KHIVA_BACKEND_CPU)
            set_backend(b)

    def test_get_device(self):
        backends = get_backends()
        cuda = backends & KHIVABackend.KHIVA_BACKEND_CUDA.value
        opencl = backends & KHIVABackend.KHIVA_BACKEND_OPENCL.value
        cpu = backends & KHIVABackend.KHIVA_BACKEND_CPU.value
        if cuda:
            set_backend(KHIVABackend.KHIVA_BACKEND_CUDA)
            for i in range(get_device_count()):
                set_device(i)
                self.assertEqual(get_device_id(), i)

        if opencl:
            set_backend(KHIVABackend.KHIVA_BACKEND_OPENCL)
            for i in range(get_device_count()):
                set_device(i)
                self.assertEqual(get_device_id(), i)

        if cpu:
            set_backend(KHIVABackend.KHIVA_BACKEND_CPU)
            for i in range(get_device_count()):
                set_device(i)
                self.assertEqual(get_device_id(), i)

    def test_version(self):
        self.assertEqual(version(), self.get_khiva_version_from_file())

    def get_khiva_version_from_github(self):
        # Hit Github API to get the list of tags.
        r = requests.get('https://api.github.com/repos/shapelets/khiva/tags')
        tag_name = '0.2.0'

        if r.status_code == 200:
            response = r.json()
            number_tags = len(response)
            tag_name = response[number_tags - 1]['name']

            # Remove symbols from numbering
            tag_name = tag_name.replace('v', '')
            tag_name = tag_name.replace('-RC', '')

        return tag_name

    def  get_khiva_version_from_file(self):

        version = ""
        if os.name == 'nt':
            path_file = "C:/Program Files/Khiva/v0/include/khiva/version.h"
        else:
            path_file = "/usr/local/include/khiva/version.h"

        version_file = open(path_file, "rt")
        contents = version_file.read()
        version_file.close()

        regex = r'([0-9]+\.[0-9]+\.[0-9]+)'
        match = re.search(regex, contents)
        if match:
           version = match.group(1)
        return version


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(LibraryTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################

def lls(a, b):
    """ Calculates the minimum norm least squares solution :math:`x` :math:`(||A·x - b||^2)` to :math:`A·x = b`. This
    function uses the singular value decomposition function of Arrayfire. The actual formula that this function computes
    is :math:`x = V·D\\dagger·U^T·b`. Where :math:`U` and :math:`V` are orthogonal matrices and :math:`D\\dagger` contains
    the inverse values of the singular values contained in :math:`D` if they are not zero, and zero otherwise.

    :param a: KHIVA array with the coefficients of the linear equation problem to solve. It accepts a list of lists or
              a numpy array with one or several time series.
    :param b: KHIVA array with the measured values.
    :return: KHIVA array with the solution to the linear equation problem minimizing the norm 2.
    """
    c = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.lls(ctypes.pointer(a.arr_reference), ctypes.pointer(b.arr_reference),
                                       ctypes.pointer(c))

    return Array(array_reference=c)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import os
from khiva.matrix import *
from khiva.array import *
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class MatrixTest(unittest.TestCase):
    DELTA = 1e-6
    DECIMAL = 6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_stomp_self_join(self):
        stomp_self_join_result = stomp_self_join(
            Array(data=[[10, 10, 11, 11, 10, 11, 10, 10, 11, 11, 10, 11, 10, 10], [
                11, 10, 10, 11, 10, 11, 11, 10, 11, 11, 10, 10, 11, 10]]), 3)
        expected_index = [6, 7, 8, 9, 10, 11, 0, 1, 2, 3, 4, 5, 9, 10, 11, 6, 7, 8, 3, 4, 5, 0, 1, 2]

        for i in range(6):
            self.assertAlmostEqual(stomp_self_join_result[0].to_numpy()[0][i], 0.0, delta=1e-2)
            self.assertEqual(stomp_self_join_result[1].to_numpy()[0][i], expected_index[i])

    def test_stomp(self):
        stomp_result = stomp(Array([[10, 11, 10, 11], [10, 11, 10, 11]]),
                             Array([[10, 11, 10, 11, 10, 11, 10, 11], [10, 11, 10, 11, 10, 11, 10, 11]]), 3)
        expected_index = [0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
        a = stomp_result[0].to_numpy().flatten()
        b = stomp_result[1].to_numpy().flatten()

        for i in range(24):
            self.assertAlmostEqual(a[i], 0, delta=1e-2)
            self.assertAlmostEqual(b[i], expected_index[i])

    def test_stomp_different_queries(self):
        stomp_result = stomp(Array([[10, 11, 10, 8, 14], [10, 14, 10, 10, 3]]),
                             Array([[10, 11, 10, 11, 10, 11, 10, 7], [10, 13, 10, 10, 10, 14, 8, 7]]), 3)

        a = stomp_result[0].to_numpy()
        b = stomp_result[1].to_numpy()

        self.assertAlmostEqual(a[1, 0, 2], 1.73205077, delta=1e-3)
        self.assertAlmostEqual(a[0, 0, 0], 0.00954336, delta=1e-3)

        self.assertEqual(b[0, 1, 5], 2)
        self.assertEqual(b[1, 1, 1], 1)

    def test_find_best_n_motifs(self):
        stomp_result = stomp(Array([10, 10, 10, 10, 10, 10, 9, 10, 10, 10, 10, 10, 11, 10, 9], dtype.f32),
                             Array([10, 11, 10, 9], dtype.f32),
                             3)

        find_best_n_motifs_result = find_best_n_motifs(stomp_result[0], stomp_result[1], 3, 1)
        a = find_best_n_motifs_result[1].to_numpy()
        b = find_best_n_motifs_result[2].to_numpy()
        self.assertAlmostEqual(a, 12, delta=self.DELTA)
        self.assertAlmostEqual(b, 1, delta=self.DELTA)

    def test_find_best_n_motifs_multiple_profiles(self):
        stomp_result = stomp(Array([[10, 10, 10, 10, 10, 10, 9, 10, 10, 10, 10, 10, 11, 10, 9],
                                    [10, 10, 10, 10, 10, 10, 9, 10, 10, 10, 10, 10, 11, 10, 9]], dtype.f32),
                             Array([[10, 11, 10, 9], [10, 11, 10, 9]], dtype.f32),
                             3)

        find_best_n_motifs_result = find_best_n_motifs(stomp_result[0], stomp_result[1], 3, 1)
        a = find_best_n_motifs_result[1].to_numpy()
        b = find_best_n_motifs_result[2].to_numpy()
        np.testing.assert_array_almost_equal(a, np.array([[[12], [12]], [[12], [12]]]), decimal=self.DECIMAL)
        np.testing.assert_array_almost_equal(b, np.array([[[1], [1]], [[1], [1]]]), decimal=self.DECIMAL)

    def test_find_best_n_motifs_mirror(self):
        stomp_result = stomp_self_join(Array([10.1, 11, 10.2, 10.15, 10.775, 10.1, 11, 10.2], dtype.f32), 3)

        find_best_n_motifs_result = find_best_n_motifs(stomp_result[0], stomp_result[1], 3, 2, True)
        a = find_best_n_motifs_result[1].to_numpy()
        b = find_best_n_motifs_result[2].to_numpy()
        self.assertAlmostEqual(a[0], 0, delta=self.DELTA)
        self.assertAlmostEqual(a[1], 0, delta=self.DELTA)
        self.assertAlmostEqual(b[0], 5, delta=self.DELTA)
        self.assertAlmostEqual(b[1], 3, delta=self.DELTA)

    def test_find_best_n_motifs_consecutive(self):
        stomp_result = stomp_self_join(Array([10.1, 11, 10.1, 10.15, 10.075, 10.1, 11, 10.1, 10.15], dtype.f32), 3)

        find_best_n_motifs_result = find_best_n_motifs(stomp_result[0], stomp_result[1], 3, 2)
        a = find_best_n_motifs_result[1].to_numpy()
        b = find_best_n_motifs_result[2].to_numpy()
        self.assertAlmostEqual(a[1], 6, delta=self.DELTA)
        self.assertAlmostEqual(b[1], 3, delta=self.DELTA)

    def test_find_best_n_discords(self):
        stomp_result = stomp(Array(np.array([11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11])),
                             Array(np.array([9, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 9])),
                             3)
        find_best_n_discords_result = find_best_n_discords(stomp_result[0],
                                                           stomp_result[1], 3, 2)
        a = find_best_n_discords_result[2].to_numpy()

        self.assertEqual(a[0], 0)
        # The test failed in the CPU used in the Travis CI build machine
        if os.environ.get("TRAVIS") == "true":
            self.assertEqual(a[1], 2)
        else:
            self.assertEqual(a[1], 10)

    def test_find_best_n_discords_multiple_profiles(self):
        stomp_result = stomp(Array(np.array([[11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11],
                                             [11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11]])),
                             Array(np.array([[9, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 9],
                                             [9, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1, 10.2, 10.1,
                                              9]])),
                             3)
        find_best_n_discords_result = find_best_n_discords(stomp_result[0],
                                                           stomp_result[1], 3, 2)
        a = find_best_n_discords_result[2].to_numpy()

        # The test failed in the CPU used in the Travis CI build machine
        if os.environ.get("TRAVIS") == "true":
            np.testing.assert_array_almost_equal(a, np.array([[[0, 2], [0, 2]], [[0, 2], [0, 2]]]),
                                                 decimal=self.DECIMAL)
        else:
            np.testing.assert_array_almost_equal(a, np.array([[[0, 10], [0, 10]], [[0, 10], [0, 10]]]),
                                                 decimal=self.DECIMAL)

    def test_find_best_n_discords_mirror(self):
        stomp_result = stomp_self_join(Array(np.array([10, 11, 10, 10, 11, 10])), 3)
        find_best_n_discords_result = find_best_n_discords(stomp_result[0],
                                                           stomp_result[1], 3, 1, True)
        a = find_best_n_discords_result[1].to_numpy()
        b = find_best_n_discords_result[2].to_numpy()

        self.assertEqual(a, 3)
        self.assertEqual(b, 1)

    def test_find_best_n_discords_consecutive(self):
        stomp_result = stomp_self_join(
            Array(np.array([10, 11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 11, 10, 9.999, 9.998]), dtype.f32), 3)
        find_best_n_discords_result = find_best_n_discords(stomp_result[0],
                                                           stomp_result[1], 3, 2, True)
        a = find_best_n_discords_result[2].to_numpy()

        self.assertEqual(a[0], 12)
        # The test failed in the CPU used in the Travis CI build machine
        if os.environ.get("TRAVIS") == "true":
            self.assertEqual(a[1], 11)
        else:
            self.assertNotEqual(a[1], 11)

    def test_mass(self):
        mass_result = mass(Array(np.array([4, 3, 8]), dtype.f32),
                           Array(np.array([10, 10, 10, 11, 12, 11, 10, 10, 11, 12, 11, 14, 10, 10]), dtype.f32))

        distances = mass_result.to_numpy()
        distances_expected = np.array([1.732051, 0.328954, 1.210135, 3.150851, 3.245858, 2.822044,
                                       0.328954, 1.210135, 3.150851, 0.248097, 3.30187, 2.82205])

        np.testing.assert_array_almost_equal(distances, distances_expected, decimal=self.DECIMAL)

    def test_mass_multiple(self):
        mass_result = mass(Array(np.array([[10, 10, 11, 11], [10, 11, 10, 10]]), dtype.f32),
                           Array(np.array([[10, 10, 10, 11, 12, 11, 10], [10, 11, 12, 11, 14, 10, 10]]), dtype.f32))

        distances = mass_result.to_numpy()

        distances_expected = np.array([[[1.83880341, 0.87391543, 1.5307337,  3.69551826],
                                        [3.26598597, 3.48967957, 2.82842779, 1.21162188]],
                                       [[1.5307337, 2.17577887, 2.57832384, 3.75498915],
                                        [2.82842779, 2.82842731, 3.21592307, 0.50202721]]])

        np.testing.assert_array_almost_equal(distances, distances_expected, decimal=self.DECIMAL)

    def test_find_best_n_occurrences(self):
        find_result = find_best_n_occurrences(
                            Array(np.array([10, 11, 12]), dtype.f32),
                            Array(np.array([[10, 10, 11, 11, 12, 11, 10, 10, 11, 12, 11, 10, 10, 11],
                                            [10, 10, 11, 11, 12, 11, 10, 10, 11, 12, 11, 10, 10, 11]]), dtype.f32),
                            1)

        distances = find_result[0].to_numpy()
        indexes = find_result[1].to_numpy()

        self.assertAlmostEqual(distances[0], 0.00069053, delta=self.DELTA)
        self.assertEqual(indexes[0], 7)

    def test_find_best_n_occurrences_multiple_queries(self):
        find_result = find_best_n_occurrences(
            Array(np.array([[11, 11, 10, 11], [10, 11, 11, 12]]), dtype.f32),
            Array(np.array([[10, 10, 11, 11, 10, 11, 10, 10, 11, 11, 10, 11, 10, 10],
                            [11, 10, 10, 11, 10, 11, 11, 10, 11, 11, 14, 10, 11, 10]]), dtype.f32),
            4)

        distances = find_result[0].to_numpy()
        indexes = find_result[1].to_numpy()

        np.testing.assert_array_equal(find_result[0].get_dims(), np.array([4, 2, 2, 1]))

        # Subsequence index of the third best distance for the second query over the first time series
        self.assertEqual(indexes[0, 1, 3], 2)

        # Second best distance for the first query over the second time series
        self.assertAlmostEqual(distances[1, 0, 2], 1.83880329, delta=self.DELTA)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(MatrixTest)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

########################################################################################################################
# IMPORT
########################################################################################################################
import ctypes
from khiva.library import KhivaLibrary
from khiva.array import Array


########################################################################################################################


def paa(a, bins):
    """ Piecewise Aggregate Approximation (PAA) approximates a time series :math:`X` of length :math:`n` into vector
    :math:`\\bar{X}=(\\bar{x}_{1},…,\\bar{x}_{M})` of any arbitrary length :math:`M \\leq n` where each of
    :math:`\\bar{x_{i}}` is calculated as follows:

    .. math::
        \\bar{x}_{i} = \\frac{M}{n} \\sum_{j=n/M(i-1)+1}^{(n/M)i} x_{j}.

    Which simply means that in order to reduce the dimensionality from :math:`n` to :math:`M`, we first divide the original
    time series into :math:`M` equally sized frames and secondly compute the mean values for each frame. The sequence
    assembled from the mean values is the PAA approximation (i.e., transform) of the original time series.

    :param a: Set of points.
    :param bins: Sets the total number of divisions.

    :return: KHIVA array of points with the reduced dimensionality.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.paa(ctypes.pointer(a.arr_reference), ctypes.pointer(ctypes.c_int(bins)),
                                       ctypes.pointer(b))

    return Array(array_reference=b)


def pip(a, number_ips):
    """ Calculates the number of Perceptually Important Points (PIP) in the time series.

    [1] Fu TC, Chung FL, Luk R, and Ng CM. Representing financial time series based on data point importance.
    Engineering Applications of Artificial Intelligence, 21(2):277-300, 2008.

    :param a: KHIVA array whose dimension zero is the length of the time series.
    :param number_ips: The number of points to be returned.

    :return: KHIVA array with the most Perceptually Important number_ips.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.pip(ctypes.pointer(a.arr_reference), ctypes.pointer(ctypes.c_int(number_ips)),
                                       ctypes.pointer(b))

    return Array(array_reference=b)


def pla_bottom_up(ts, max_error):
    """
    Applies the Piecewise Linear Approximation (PLA BottomUP) to the time series.

    [1] Zhu Y, Wu D, Li Sh (2007). A Piecewise Linear Representation Method of Time Series Based on Feature Points.
    Knowledge-Based Intelligent Information and Engineering Systems 4693:1066-1072.

    :param ts: Expects a khiva array containing the set of points to be reduced. The first component of the points in
                the first column and the second component of the points in the second column.
    :param max_error: The maximum approximation error allowed.
    :return: The reduced number of points.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.pla_bottom_up(ctypes.pointer(ts.arr_reference),
                                                 ctypes.pointer(ctypes.c_float(max_error)),
                                                 ctypes.pointer(b))

    return Array(array_reference=b)


def pla_sliding_window(ts, max_error):
    """
    Applies the Piecewise Linear Approximation (PLA Sliding Window) to the time series.

    [1] Zhu Y, Wu D, Li Sh (2007). A Piecewise Linear Representation Method of Time Series Based on Feature Points.
    Knowledge-Based Intelligent Information and Engineering Systems 4693:1066-1072.

    :param ts: Expects a khiva array containing the set of points to be reduced. The first component of the points in
                the first column and the second component of the points in the second column.
    :param max_error: The maximum approximation error allowed.
    :return: The reduced number of points.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.pla_sliding_window(ctypes.pointer(ts.arr_reference),
                                                      ctypes.pointer(ctypes.c_float(max_error)),
                                                      ctypes.pointer(b))

    return Array(array_reference=b)


def ramer_douglas_peucker(a, epsilon):
    """ The Ramer–Douglas–Peucker algorithm (RDP) is an algorithm for reducing the number of points in a curve that is
    approximated by a series of points. It reduces a set of points depending on the perpendicular distance of the points
    and epsilon, the greater epsilon, more points are deleted.

    [1] Urs Ramer, "An iterative procedure for the polygonal approximation of plane curves", Computer Graphics and Image
    Processing, 1(3), 244–256 (1972) doi:10.1016/S0146-664X(72)80017-0.

    [2] David Douglas & Thomas Peucker, "Algorithms for the reduction of the number of points required to represent a
    digitized line or its caricature", The Canadian Cartographer 10(2), 112–122 (1973) doi:10.3138/FM57-6770-U75U-7727

    :param a: KHIVA array with the x-coordinates and y-coordinates of the input points (x in column 0 and y in column 1).
    :param epsilon: It acts as the threshold value to decide which points should be considered meaningful or not.

    :return: KHIVA array with the x-coordinates and y-coordinates of the selected points (x in column 0 and y in
            column 1).
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.ramer_douglas_peucker(ctypes.pointer(a.arr_reference),
                                                         ctypes.pointer(ctypes.c_double(epsilon)),
                                                         ctypes.pointer(b))

    return Array(array_reference=b)


def sax(a, alphabet_size):
    """ Symbolic Aggregate approXimation (SAX). It transforms a numeric time series into a time series of symbols with
    the same size. The algorithm was proposed by Lin et al.) and extends the PAA-based approach inheriting the original
    algorithm simplicity and low computational complexity while providing satisfactory sensitivity and selectivity in
    range query processing. Moreover, the use of a symbolic representation opened a door to the existing wealth of
    data-structures and string-manipulation algorithms in computer science such as hashing, regular expression, pattern
    matching, suffix trees, and grammatical inference.

    [1] Lin, J., Keogh, E., Lonardi, S. & Chiu, B. (2003) A Symbolic Representation of Time Series, with Implications
    for Streaming Algorithms. In proceedings of the 8th ACM SIGMOD Workshop on Research Issues in Data Mining and
    Knowledge Discovery. San Diego, CA. June 13.

    :param a: KHIVA array with the input time series.
    :param alphabet_size: Number of element within the alphabet.

    :return: KHIVA array of points with the reduced dimensionality.
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.sax(ctypes.pointer(a.arr_reference), ctypes.pointer(ctypes.c_int(alphabet_size)),
                                       ctypes.pointer(b))

    return Array(array_reference=b)


def visvalingam(a, num_points):
    """ Reduces a set of points by applying the Visvalingam method (minimun triangle area) until the number
    of points is reduced to numPoints.

    [1] M. Visvalingam and J. D. Whyatt, Line generalisation by repeated elimination of points,
    The Cartographic Journal, 1993.

    :param a: KHIVA array with the x-coordinates and y-coordinates of the input points (x in column 0 and y in column 1).
    :param num_points: Sets the number of points returned after the execution of the method.

    :return: KHIVA array with the x-coordinates and y-coordinates of the selected points (x in column 0 and y in
            column 1).
    """
    b = ctypes.c_void_p(0)
    KhivaLibrary().c_khiva_library.visvalingam(ctypes.pointer(a.arr_reference),
                                               ctypes.pointer(ctypes.c_int(num_points)),
                                               ctypes.pointer(b))

    return Array(array_reference=b)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Shapelets.io
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.


########################################################################################################################
# IMPORT
########################################################################################################################
import unittest
import numpy as np
from khiva.regularization import *
from khiva.array import Array
from khiva.library import set_backend, KHIVABackend


########################################################################################################################

class RegularizationTest(unittest.TestCase):
    DELTA = 1e-6
    DECIMAL = 6

    def setUp(self):
        set_backend(KHIVABackend.KHIVA_BACKEND_CPU)

    def test_group_by_single_column(self):
        group_by_result = group_by(Array([[0, 1, 1, 2, 2, 3], [0, 3, 3, 1, 1, 2]]), 0).to_numpy()
        expected = [0, 3, 1, 2]
        np.testing.assert_array_almost_equal(group_by_result, expected, decimal=self.DECIMAL)

    def test_group_by_double_key_column(self):
        group_by_result = group_by(Array([[0, 1, 1, 2, 2, 3], [1, 2, 2, 3, 3, 4], [0, 3, 3, 1, 1, 2]]),
                                   0, 2, 1).to_numpy()
        expected = [0, 3, 1, 2]
        np.testing.assert_array_almost_equal(group_by_result, expected, decimal=self.DECIMAL)

    def test_group_by_double_key_column_2(self):
        group_by_result = group_by(Array([[0, 0, 1, 1, 1], [0, 1, 0, 0, 1], [1, 2, 3, 4, 5]]),
                                   0, 2, 1).to_numpy()
        expected = [1, 2, 3.5, 5]
        np.testing.assert_array_almost_equal(group_by_result, expected, decimal=self.DECIMAL)

    def test_group_by_double_key_double_value_column(self):
        group_by_result = group_by(Array([[0, 0, 0, 2, 2], [2, 2, 2, 4, 4], [0, 1, 2, 3, 4], [1, 1, 1, 1, 1]]),
                                   0, 2, 2).to_numpy()
        expected = [[1, 3.5], [1, 1]]
        np.testing.assert_array_almost_equal(group_by_result, expected, decimal=self.DECIMAL)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(RegularizationTest)
    unittest.TextTestRunner(verbosity=2).run(suite)
